import os, sys, traceback
from os.path import join, exists, isdir, isfile, dirname, abspath, expanduser
import datetime
from gi.repository import Gio
import pickle
from urllib.request import ProxyHandler, build_opener, install_opener
from . import networkmanager

# Autotools set the actual data_dir in defs.py
from .defs import DATA_DIR, BUILDERDIR

DEBUGGING = False

# central debugging and error method
def debug(msg):
	if DEBUGGING:
		print("%s: %s" % (datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"), msg))

def error(msg):
	print("%s: ERROR: %s" % (datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"), msg))


def exceptionhandler(t, value, tb):
	error("%s occurred: %s" % (t.__name__, value))
	for elem in traceback.extract_tb(tb):
		error("\t%s:%d in %s: %s" % (elem[0], elem[1], elem[2], elem[3]))

# redirect uncaught exceptions to this exception handler
debug("Installing default exception handler")
sys.excepthook = exceptionhandler



# Allow to use uninstalled invest ---------------------------------------------
UNINSTALLED_INVEST = False
def _check(path):
	return exists(path) and isdir(path) and isfile(path+"/Makefile.am")

name = join(dirname(__file__), '..')
if _check(name):
	UNINSTALLED_INVEST = True

# Sets SHARED_DATA_DIR to local copy, or the system location
# Shared data dir is most the time /usr/share/invest-applet
if UNINSTALLED_INVEST:
	SHARED_DATA_DIR = abspath(join(dirname(__file__), '..', 'data'))
	BUILDER_DATA_DIR = SHARED_DATA_DIR
	ART_DATA_DIR = join(SHARED_DATA_DIR, 'art')
else:
	SHARED_DATA_DIR = join(DATA_DIR, "gnome-applets", "invest-applet")
	BUILDER_DATA_DIR = BUILDERDIR
	ART_DATA_DIR = SHARED_DATA_DIR

USER_INVEST_DIR = expanduser("~/.config/gnome-applets/invest-applet")
if not exists(USER_INVEST_DIR):
	try:
		os.makedirs(USER_INVEST_DIR, 0o744)
	except Exception as msg:
		error('Could not create user dir (%s): %s' % (USER_INVEST_DIR, msg))
# ------------------------------------------------------------------------------

# Set the cwd to the home directory so spawned processes behave correctly
# when presenting save/open dialogs
os.chdir(expanduser("~"))

# tests whether the given stocks are in the old labelless format
def labelless_stock_format(stocks):
	if len(stocks) == 0:
		return False

	# take the first element of the dict and check if its value is a list
	if type(stocks[list(stocks.keys())[0]]) is list:
		return True

	# there is no list, so it is already the new stock file format
	return False

# converts the given stocks from the labelless format into the one with labels
def update_to_labeled_stock_format(stocks):
	new = {}

	for k, l in stocks.items():
		d = {'label':"", 'purchases':l}
		new[k] = d

	return new

# tests whether the given stocks are in the format without exchange information
def exchangeless_stock_format(stocks):
	if len(stocks) == 0:
		return False

	# take the first element of the dict and check if its value is a list
	for symbol, data in stocks.items():
		purchases = stocks[symbol]["purchases"]
		if len(purchases) > 0:
			purchase = purchases[0]
			if "exchange" not in purchase:
				return True

	return False

# converts the given stocks into format with exchange information
def update_to_exchange_stock_format(stocks):
	for symbol, data in stocks.items():
		purchases = data["purchases"]
		for purchase in purchases:
			purchase["exchange"] = 0

	return stocks

# converts the given stocks from the dict format into a list
def update_to_list_stock_format(stocks):
	new = []

	for ticker, stock in stocks.items():
		stock['ticker'] = ticker
		new.append(stock)

	return new

STOCKS_FILE = join(USER_INVEST_DIR, "stocks.pickle")

try:
	with open(STOCKS_FILE, 'rb') as stocks_file:
		STOCKS = pickle.load(stocks_file)

	# if the stocks file contains a list, the subsequent tests are obsolete
	if type(STOCKS) != list:
		# if the stocks file is in the stocks format without labels,
		# then we need to convert it into the new labeled format
		if labelless_stock_format(STOCKS):
			STOCKS = update_to_labeled_stock_format(STOCKS)

		# if the stocks file does not contain exchange rates, add them
		if exchangeless_stock_format(STOCKS):
			STOCKS = update_to_exchange_stock_format(STOCKS)

		# here, stocks is a most up-to-date dict, but we need it to be a list
		STOCKS = update_to_list_stock_format(STOCKS)

except Exception as msg:
	error("Could not load the stocks from %s: %s" % (STOCKS_FILE, msg) )
	STOCKS = []

#STOCKS = {
#	"AAPL": {
#		"amount": 12,
#		"bought": 74.94,
#		"comission": 31,
#	},
#	"INTC": {
#		"amount": 30,
#		"bought": 25.85,
#		"comission": 31,
#	},
#	"GOOG": {
#		"amount": 1,
#		"bought": 441.4,
#		"comission": 31,
#	},
#}

CONFIG_FILE = join(USER_INVEST_DIR, "config.pickle")
try:
	with open(CONFIG_FILE, 'rb') as config_file:
		CONFIG = pickle.load(config_file)
except Exception as msg:
	error("Could not load the configuration from %s: %s" % (CONFIG_FILE, msg) )
	CONFIG = {}       # default configuration

CURRENCIES_FILE = join(USER_INVEST_DIR, "currencies.csv")
QUOTES_FILE = join(USER_INVEST_DIR, "quotes.csv")
INDEX_QUOTES_FILE_TEMPLATE = "quotes.#.csv"


# set default proxy config
PROXY = None

# borrowed from Ross Burton
# http://burtonini.com/blog/computers/postr
# extended by exception handling and retry scheduling
def get_gnome_proxy():
	# sanity check if we still need to look for proxy configuration
	global PROXY
	if PROXY != None:
		return

	# try to get config from gsettings
	try:
		proxy_settings = Gio.Settings.new("org.gnome.system.proxy")
		proxy_http_settings = Gio.Settings.new("org.gnome.system.proxy.http")

		if proxy_settings.get_enum("mode") == 1:
			host = proxy_http_settings.get_string("host")
			port = proxy_http_settings.get_int("port")
			if host is None or host == "" or port == 0:
				# gnome proxy is not valid, stop here
				return

			if proxy_http_settings.get_boolean("use-authentication"):
				user = proxy_http_settings.get_string("authentication-user")
				password = proxy_http_settings.get_string("authentication-password")
				if user and user != "":
					url = "http://%s:%s@%s:%d" % (user, password, host, port)
				else:
					url = "http://%s:%d" % (host, port)
			else:
				url = "http://%s:%d" % (host, port)

			# proxy config found, memorize
			PROXY = {'http': url}
			proxy_handler = ProxyHandler(proxies=PROXY)
			install_opener(build_opener(proxy_handler))

	except Exception as msg:
		error("Failed to get proxy configuration from GSettings:\n%s" % msg)

# use gsettings to get proxy config
debug("Detecting proxy settings")
get_gnome_proxy()


# connect to Network Manager to identify current network connectivity
debug("Connecting to the NetworkManager")
nm = networkmanager.NetworkManager()



from os.path import join

from gi.repository import GObject, Gio, Gtk, Gdk, GdkPixbuf, PanelApplet
GObject.threads_init()

from gettext import gettext as _

import invest, invest.about, invest.chart, invest.preferences, invest.defs
from invest.quotes import QuoteUpdater
from invest.widgets import InvestWidget

Gtk.Window.set_default_icon_from_file(join(invest.ART_DATA_DIR, "invest_neutral.svg"))

WidgetBorderWidth = 4
WidgetPaddingSpace = 10

class InvestApplet(PanelApplet.Applet):
	def __init__(self, applet):
		invest.debug("init applet");
		self.applet = applet

		actiongroup = Gio.SimpleActionGroup.new()

		action = Gio.SimpleAction.new("about", None)
		action.connect("activate", self.on_about)
		actiongroup.add_action(action)

		action = Gio.SimpleAction.new("help", None)
		action.connect("activate", self.on_help)
		actiongroup.add_action(action)

		action = Gio.SimpleAction.new("preferences", None)
		action.connect("activate", self.on_preferences)
		actiongroup.add_action(action)

		action = Gio.SimpleAction.new("refresh", None)
		action.connect("activate", self.on_refresh)
		actiongroup.add_action(action)

		self.applet.setup_menu_from_file (join(invest.defs.PKGDATADIR, "ui/invest-applet-menu.xml"),
						  actiongroup, invest.defs.GETTEXT_PACKAGE);

		self.applet.insert_action_group ("invest", actiongroup);

		evbox = Gtk.HBox()
		self.applet_icon = Gtk.Image()
		self.set_applet_icon(0)
		self.applet_icon.show()
		evbox.add(self.applet_icon)
		self.applet.add(evbox)
		self.applet.connect("button-press-event", self.button_clicked)
		self.applet.show_all()
		self.new_ilw()

	def new_ilw(self):
		self.quotes_updater = QuoteUpdater(self.set_applet_icon,
						   self.set_applet_tooltip)
		self.investwidget = InvestWidget(self.quotes_updater)
		self.ilw = InvestmentsListWindow(self.applet, self.investwidget)

	def reload_ilw(self):
		self.ilw.destroy()
		self.new_ilw()

	def button_clicked(self, widget,event):
		if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
			# Three cases...
			if len (invest.STOCKS) == 0:
				# a) We aren't configured yet
				invest.preferences.show_preferences(self, _("<b>You have not entered any stock information yet</b>"))
				self.reload_ilw()
			elif not self.quotes_updater.quotes_valid:
				# b) We can't get the data (e.g. offline)
				alert = Gtk.MessageDialog(buttons=Gtk.ButtonsType.CLOSE)
				alert.set_markup(_("<b>No stock quotes are currently available</b>"))
				alert.format_secondary_text(_("The server could not be contacted. The computer is either offline or the servers are down. Try again later."))
				alert.run()
				alert.destroy()
			else:
				# c) Everything is normal: pop-up the window
				self.ilw.toggle_show()
	
	def on_about(self, action, parameter):
		invest.about.show_about()
	
	def on_help(self, action, parameter):
		invest.help.show_help()

	def on_preferences(self, action, parameter):
		invest.preferences.show_preferences(self)
		self.reload_ilw()
	
	def on_refresh(self, action, parameter):
		self.quotes_updater.refresh()

	def set_applet_icon(self, change):
		if change == 1:
			pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(join(invest.ART_DATA_DIR, "invest-22_up.png"), -1,-1)
		elif change == 0:
			pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(join(invest.ART_DATA_DIR, "invest-22_neutral.png"), -1,-1)
		else:
			pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(join(invest.ART_DATA_DIR, "invest-22_down.png"), -1,-1)
		self.applet_icon.set_from_pixbuf(pixbuf)
	
	def set_applet_tooltip(self, text):
		self.applet_icon.set_tooltip_text(text)

class InvestmentsListWindow(Gtk.Window):
	def __init__(self, applet, list):
		Gtk.Window.__init__(self, type=Gtk.WindowType.TOPLEVEL)
		self.list = list
		self.list.connect('row-collapsed', self.on_collapsed)
		self.list.connect('row-expanded', self.on_expanded)

		self.set_type_hint(Gdk.WindowTypeHint.DOCK)
		self.stick()
		self.set_border_width(WidgetBorderWidth)
		
		self.applet = applet # this is the widget we want to align with
		self.alignment = self.applet.get_orient ()

		sw = Gtk.ScrolledWindow()
		sw.set_policy(Gtk.PolicyType.NEVER,Gtk.PolicyType.AUTOMATIC)
		sw.add(list)
		self.add(sw)

		# boolean variable that identifies if the window is visible
		# show/hide is triggered by left-clicking on the applet
		self.hidden = True

	def toggle_show(self):
		if self.hidden == True:
			self.update_position()
			self.show_all()
			self.hidden = False
		elif self.hidden == False:
			self.hide()
			self.hidden = True

	def on_collapsed(self, treeview, iter, path):
		self.update_position()

	def on_expanded(self, treeview, iter, path):
		self.update_position()

	def update_position (self):
		"""
		Calculates the position and moves the window to it.
		"""
		self.realize()

		# get information sources
		window = self.applet.get_window()
		screen = window.get_screen()
		monitor = screen.get_monitor_geometry (screen.get_monitor_at_window (window))

		# Get applet position and dimensions ...
		(ret, ax, ay) = window.get_origin()
		(ignored, ignored, aw, ah) = window.get_geometry()

		# ... and widget dimensions
		size = self.list.size_request()
		ww = size.width + self.get_border_width() * 2
		wh = size.height + self.get_border_width() * 2

		invest.debug("applet has position (%d,%d) and size (%d,%d)" % (ax, ay, aw, ah))
		invest.debug("applet widget has size (%d,%d)" % (ww, wh))
		invest.debug("monitor has position (%d,%d) and size (%d,%d)" % (monitor.x, monitor.y, monitor.width, monitor.height))
		if self.alignment == PanelApplet.AppletOrient.LEFT:
			invest.debug("applet orientation is LEFT")
			wx = ax - ww - WidgetPaddingSpace
			wy = ay

			if (wy + wh + WidgetPaddingSpace > monitor.y + monitor.height):
				wy = monitor.y + monitor.height - wh - WidgetPaddingSpace

			if (wy + wh + WidgetPaddingSpace > monitor.height):
				wh = wh - (wy + wh + WidgetPaddingSpace - monitor.height)
				wy = monitor.height - wh - WidgetPaddingSpace

			if (wy < WidgetPaddingSpace):
				wh = wh + (wy - WidgetPaddingSpace)
				wy = WidgetPaddingSpace

			if (wy + wh > monitor.height / 2):
				gravity = Gdk.Gravity.SOUTH_WEST
			else:
				gravity = Gdk.Gravity.NORTH_WEST

		elif self.alignment == PanelApplet.AppletOrient.RIGHT:
			invest.debug("applet orientation is RIGHT")
			wx = ax + aw
			wy = ay

			if (wy + wh + WidgetPaddingSpace > monitor.y + monitor.height):
				wy = monitor.y + monitor.height - wh - WidgetPaddingSpace

			if (wy + wh + WidgetPaddingSpace > monitor.height):
				wh = wh - (wy + wh + WidgetPaddingSpace - monitor.height)
				wy = monitor.height - wh - WidgetPaddingSpace

			if (wy < WidgetPaddingSpace):
				wh = wh + (wy - WidgetPaddingSpace)
				wy = WidgetPaddingSpace

			if (wy + wh > monitor.height / 2):
				gravity = Gdk.Gravity.SOUTH_EAST
			else:
				gravity = Gdk.Gravity.NORTH_EAST

		elif self.alignment == PanelApplet.AppletOrient.DOWN:
			invest.debug("applet orientation is DOWN")
			wx = ax
			wy = ay + ah

			if (wx + ww > monitor.x + monitor.width):
				wx = monitor.x + monitor.width - ww
			if (wx < 0):
				wx = 0

			if (wy + wh + WidgetPaddingSpace > monitor.height):
				wh = wh - (wy + wh + WidgetPaddingSpace - monitor.height)
				wy = monitor.height - wh - WidgetPaddingSpace

			gravity = Gdk.Gravity.NORTH_WEST
		elif self.alignment == PanelApplet.AppletOrient.UP:
			invest.debug("applet orientation is UP")
			invest.debug("widget size is (%d,%d)" % (ww, wh))
			wx = ax
			wy = ay - wh

			if (wx + ww > monitor.x + monitor.width):
				wx = monitor.x + monitor.width - ww
			if (wx < 0):
				wx = 0

			invest.debug("current widget position is (%d,%d)" % (wx, wy))
			if (wy < WidgetPaddingSpace):
				wy = WidgetPaddingSpace
			if (wy + wh + WidgetPaddingSpace > monitor.height):
				wh = monitor.height - wy - WidgetPaddingSpace


			gravity = Gdk.Gravity.SOUTH_WEST

		invest.debug("Set widget position to (%d,%d)" % (wx, wy))
		self.move(wx, wy)
		invest.debug("Set widget size to (%d,%d)" % (ww, wh))
		self.resize(ww, wh)
		invest.debug("Set widget gravity to %s" % gravity)
		self.set_gravity(gravity)
